# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of seq2seq model.

The model is based on: https://github.com/google/flax/tree/main/examples/seq2seq
"""

import logging
import numpy as np
from typing import Tuple
from typing import Any
from functools import partial

import jax
import jax.numpy as jnp
from jax import random
from flax import linen as nn

from evojax.policy.base import PolicyNetwork
from evojax.policy.base import PolicyState
from evojax.task.base import TaskState
from evojax.util import create_logger
from evojax.util import get_params_format_fn


class CharacterTable(object):
    """Encode/decodes between strings and integer representations."""

    def __init__(self):
        self._chars = '0123456789+= '
        self.pad_id = len(self._chars)
        self.eos_id = self.pad_id + 1
        self.vocab_size = len(self._chars) + 2
        self._indices_char = dict(
            (idx, ch) for idx, ch in enumerate(self._chars))
        self._indices_char[self.pad_id] = '_'

    def encode(self, inputs: jnp.ndarray) -> jnp.ndarray:
        return jnp.concatenate([inputs, jnp.array([self.eos_id])])

    def decode(self, inputs):
        """Decode from list of integers to string."""
        chars = []
        for elem in inputs.tolist():
            if elem == self.eos_id:
                break
            chars.append(self._indices_char[elem])
        return ''.join(chars)


char_table = CharacterTable()


class EncoderLSTM(nn.Module):
    """LSTM in the encoder part of the seq2seq model."""

    @partial(
        nn.transforms.scan,
        variable_broadcast='params',
        in_axes=1,
        out_axes=1,
        split_rngs={'params': False})
    @nn.compact
    def __call__(self, carry, x):
        lstm_state, is_eos = carry
        new_lstm_state, y = nn.LSTMCell()(lstm_state, x)

        # Pass forward the previous state if EOS has already been reached.
        def select_carried_state(new_state, old_state):
            return jnp.where(is_eos[:, np.newaxis], old_state, new_state)

        # LSTM state is a tuple (c, h).
        carried_lstm_state = tuple(
            select_carried_state(*s) for s in zip(new_lstm_state, lstm_state))
        # Update `is_eos`.
        is_eos = jnp.logical_or(is_eos, x[:, char_table.eos_id])
        return (carried_lstm_state, is_eos), y

    @staticmethod
    def initialize_carry(batch_size, hidden_size):
        # use dummy key since default state init fn is just zeros.
        return nn.LSTMCell.initialize_carry(
            jax.random.PRNGKey(0), (batch_size,), hidden_size)


class Encoder(nn.Module):
    """LSTM encoder, returning state after EOS is input."""

    hidden_size: int

    @nn.compact
    def __call__(self, inputs):
        # inputs.shape = (batch_size, seq_length, vocab_size).
        batch_size = inputs.shape[0]
        lstm = EncoderLSTM(name='encoder_lstm')
        init_lstm_state = lstm.initialize_carry(batch_size, self.hidden_size)
        init_is_eos = jnp.zeros(batch_size, dtype=np.bool)
        init_carry = (init_lstm_state, init_is_eos)
        (final_state, _), _ = lstm(init_carry, inputs)
        return final_state


class DecoderLSTM(nn.Module):
    """LSTM in the decoder part of the seq2seq model."""

    teacher_force: bool

    @partial(
        nn.transforms.scan,
        variable_broadcast='params',
        in_axes=1,
        out_axes=1,
        split_rngs={'params': False})
    @nn.compact
    def __call__(self, carry, x):
        lstm_state, last_prediction = carry
        if not self.teacher_force:
            x = last_prediction
        lstm_state, y = nn.LSTMCell()(lstm_state, x)
        logits = nn.Dense(features=char_table.vocab_size)(y)
        predicted_token = jnp.argmax(logits, axis=-1)
        prediction = jax.nn.one_hot(
            predicted_token, char_table.vocab_size, dtype=jnp.float32)

        return (lstm_state, prediction), (logits, prediction)


class Decoder(nn.Module):
    """LSTM decoder."""

    init_state: Tuple[Any]
    teacher_force: bool

    @nn.compact
    def __call__(self, inputs):
        # inputs.shape = (seq_length, vocab_size).
        lstm = DecoderLSTM(teacher_force=self.teacher_force)
        init_carry = (self.init_state, inputs[:, 0])
        _, (logits, predictions) = lstm(init_carry, inputs)
        return logits, predictions


class Seq2seq(nn.Module):
    """Sequence-to-sequence class using encoder/decoder architecture."""

    teacher_force: bool
    hidden_size: int

    @nn.compact
    def __call__(self, encoder_inputs, decoder_inputs):
        # Encode inputs.
        init_decoder_state = Encoder(
            hidden_size=self.hidden_size)(encoder_inputs)
        # Decode outputs.
        logits, predictions = Decoder(
            init_state=init_decoder_state,
            teacher_force=self.teacher_force)(decoder_inputs[:, :-1])
        return logits, predictions


class Seq2seqPolicy(PolicyNetwork):
    """A seq2seq policy that deals with simple additions."""

    def __init__(self,
                 hidden_size: int = 256,
                 teacher_force: bool = False,
                 max_len_query_digit: int = 3,
                 logger: logging.Logger = None):
        if logger is None:
            self._logger = create_logger('Seq2seqPolicy')
        else:
            self._logger = logger

        max_input_len = max_len_query_digit + 2 + 2
        max_output_len = max_len_query_digit + 3
        encoder_shape = jnp.ones(
            (1, max_input_len, char_table.vocab_size), dtype=jnp.float32)
        decoder_shape = jnp.ones(
            (1, max_output_len, char_table.vocab_size), dtype=jnp.float32)
        model = Seq2seq(hidden_size=hidden_size, teacher_force=teacher_force)
        key = random.PRNGKey(0)
        params = model.init({'params': key, 'lstm': key},
                            encoder_shape, decoder_shape)['params']
        self.num_params, format_params_fn = get_params_format_fn(params)
        self._logger.info(
            'Seq2seqPolicy.num_params = {}'.format(self.num_params))
        self._format_params_fn = jax.vmap(format_params_fn)

        def forward_fn(p, o):
            x = jax.nn.one_hot(
                char_table.encode(jnp.array([11]))[0:1], char_table.vocab_size,
                dtype=jnp.float32)
            x = jnp.tile(x, (o.shape[0], max_output_len, 1))
            logits, predictions = model.apply({'params': p}, o, x)
            return logits
        self._forward_fn = jax.vmap(forward_fn)

    def get_actions(self,
                    t_states: TaskState,
                    params: jnp.ndarray,
                    p_states: PolicyState) -> Tuple[jnp.ndarray, PolicyState]:
        params = self._format_params_fn(params)
        return self._forward_fn(params, t_states.obs), p_states



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from functools import partial
from typing import Tuple
from typing import Union

import jax
import jax.numpy as jnp
from jax import random
from jax.tree_util import tree_map

from evojax.obs_norm import ObsNormalizer
from evojax.task.base import TaskState
from evojax.task.base import VectorizedTask
from evojax.policy.base import PolicyState
from evojax.policy.base import PolicyNetwork
from evojax.util import create_logger


@partial(jax.jit, static_argnums=(1, 2, 3, 4, 5))
def get_task_reset_keys(key: jnp.ndarray,
                        test: bool,
                        pop_size: int,
                        n_tests: int,
                        n_repeats: int,
                        ma_training: bool) -> Tuple[jnp.ndarray, jnp.ndarray]:
    key, subkey = random.split(key=key)
    if ma_training:
        reset_keys = random.split(subkey, n_repeats)
    else:
        if test:
            reset_keys = random.split(subkey, n_tests * n_repeats)
        else:
            reset_keys = random.split(subkey, n_repeats)
            reset_keys = jnp.tile(reset_keys, (pop_size, 1))
    return key, reset_keys


@jax.jit
def split_params_for_pmap(param: jnp.ndarray) -> jnp.ndarray:
    return jnp.stack(jnp.split(param, jax.local_device_count()))


@jax.jit
def split_states_for_pmap(
        state: Union[TaskState, PolicyState]) -> Union[TaskState, PolicyState]:
    return tree_map(split_params_for_pmap, state)


@jax.jit
def reshape_data_from_pmap(data: jnp.ndarray) -> jnp.ndarray:
    # data.shape = (#device, steps, #jobs/device, *)
    data = data.transpose([1, 0] + [i for i in range(2, data.ndim)])
    return jnp.reshape(data, (data.shape[0], data.shape[1] * data.shape[2], -1))


@partial(jax.jit, static_argnums=(1, 2))
def duplicate_params(params: jnp.ndarray,
                     repeats: int,
                     ma_training: bool) -> jnp.ndarray:
    if ma_training:
        return jnp.tile(params, (repeats, ) + (1,) * (params.ndim - 1))
    else:
        return jnp.repeat(params, repeats=repeats, axis=0)


class SimManager(object):
    """Simulation manager."""

    def __init__(self,
                 n_repeats: int,
                 test_n_repeats: int,
                 pop_size: int,
                 n_evaluations: int,
                 policy_net: PolicyNetwork,
                 train_vec_task: VectorizedTask,
                 valid_vec_task: VectorizedTask,
                 seed: int = 0,
                 obs_normalizer: ObsNormalizer = None,
                 logger: logging.Logger = None):
        """Initialization function.

        Args:
            n_repeats - Number of repeated parameter evaluations.
            pop_size - Population size.
            n_evaluations - Number of evaluations of the best parameter.
            policy_net - Policy network.
            train_vec_task - Vectorized tasks for training.
            valid_vec_task - Vectorized tasks for validation.
            seed - Random seed.
            obs_normalizer - Observation normalization helper.
            logger - Logger.
        """

        if logger is None:
            self._logger = create_logger(name='SimManager')
        else:
            self._logger = logger

        self._key = random.PRNGKey(seed=seed)
        self._n_repeats = n_repeats
        self._test_n_repeats = test_n_repeats
        self._pop_size = pop_size
        self._n_evaluations = max(n_evaluations, jax.local_device_count())
        self._ma_training = train_vec_task.multi_agent_training

        self.obs_normalizer = obs_normalizer
        if self.obs_normalizer is None:
            self.obs_normalizer = ObsNormalizer(
                obs_shape=train_vec_task.obs_shape,
                dummy=True,
            )
        self.obs_params = self.obs_normalizer.get_init_params()

        self._num_device = jax.local_device_count()
        if self._pop_size % self._num_device != 0:
            raise ValueError(
                'pop_size must be multiples of GPU/TPUs: '
                'pop_size={}, #devices={}'.format(
                    self._pop_size, self._num_device))
        if self._n_evaluations % self._num_device != 0:
            raise ValueError(
                'n_evaluations must be multiples of GPU/TPUs: '
                'n_evaluations={}, #devices={}'.format(
                    self._n_evaluations, self._num_device))

        def step_once(carry, input_data, task):
            (task_state, policy_state, params, obs_params,
             accumulated_reward, valid_mask) = carry
            if task.multi_agent_training:
                num_tasks, num_agents = task_state.obs.shape[:2]
                task_state = task_state.replace(
                    obs=task_state.obs.reshape((-1, *task_state.obs.shape[2:])))
            org_obs = task_state.obs
            normed_obs = self.obs_normalizer.normalize_obs(org_obs, obs_params)
            task_state = task_state.replace(obs=normed_obs)
            actions, policy_state = policy_net.get_actions(
                task_state, params, policy_state)
            if task.multi_agent_training:
                task_state = task_state.replace(
                    obs=task_state.obs.reshape(
                        (num_tasks, num_agents, *task_state.obs.shape[1:])))
                actions = actions.reshape(
                    (num_tasks, num_agents, *actions.shape[1:]))
            task_state, reward, done = task.step(task_state, actions)
            if task.multi_agent_training:
                reward = reward.ravel()
                done = jnp.repeat(done, num_agents, axis=0)
            accumulated_reward = accumulated_reward + reward * valid_mask
            valid_mask = valid_mask * (1 - done.ravel())
            return ((task_state, policy_state, params, obs_params,
                     accumulated_reward, valid_mask),
                    (org_obs, valid_mask))

        def rollout(task_states, policy_states, params, obs_params,
                    step_once_fn, max_steps):
            accumulated_rewards = jnp.zeros(params.shape[0])
            valid_masks = jnp.ones(params.shape[0])
            ((task_states, policy_states, params, obs_params,
              accumulated_rewards, valid_masks),
             (obs_set, obs_mask)) = jax.lax.scan(
                step_once_fn,
                (task_states, policy_states, params, obs_params,
                 accumulated_rewards, valid_masks), (), max_steps)
            return accumulated_rewards, obs_set, obs_mask

        self._policy_reset_fn = jax.jit(policy_net.reset)

        # Set up training functions.
        self._train_reset_fn = train_vec_task.reset
        self._train_rollout_fn = partial(
            rollout,
            step_once_fn=partial(step_once, task=train_vec_task),
            max_steps=train_vec_task.max_steps)
        if self._num_device > 1:
            self._train_rollout_fn = jax.jit(jax.pmap(
                self._train_rollout_fn, in_axes=(0, 0, 0, None)))

        # Set up validation functions.
        self._valid_reset_fn = valid_vec_task.reset
        self._valid_rollout_fn = partial(
            rollout,
            step_once_fn=partial(step_once, task=valid_vec_task),
            max_steps=valid_vec_task.max_steps)
        if self._num_device > 1:
            self._valid_rollout_fn = jax.jit(jax.pmap(
                self._valid_rollout_fn, in_axes=(0, 0, 0, None)))

    def eval_params(self, params: jnp.ndarray, test: bool) -> jnp.ndarray:
        """Evaluate population parameters or test the best parameter.

        Args:
            params - Parameters to be evaluated.
            test - Whether we are testing the best parameter
        Returns:
            An array of fitness scores.
        """

        policy_reset_func = self._policy_reset_fn
        if test:
            n_repeats = self._test_n_repeats
            task_reset_func = self._valid_reset_fn
            rollout_func = self._valid_rollout_fn
            params = duplicate_params(
                params[None, :], self._n_evaluations, False)
        else:
            n_repeats = self._n_repeats
            task_reset_func = self._train_reset_fn
            rollout_func = self._train_rollout_fn

        # Suppose pop_size=2 and n_repeats=3.
        # For multi-agents training, params become
        #   a1, a2, ..., an  (individual 1 params)
        #   b1, b2, ..., bn  (individual 2 params)
        #   a1, a2, ..., an  (individual 1 params)
        #   b1, b2, ..., bn  (individual 2 params)
        #   a1, a2, ..., an  (individual 1 params)
        #   b1, b2, ..., bn  (individual 2 params)
        # For non-ma training, params become
        #   a1, a2, ..., an  (individual 1 params)
        #   a1, a2, ..., an  (individual 1 params)
        #   a1, a2, ..., an  (individual 1 params)
        #   b1, b2, ..., bn  (individual 2 params)
        #   b1, b2, ..., bn  (individual 2 params)
        #   b1, b2, ..., bn  (individual 2 params)
        params = duplicate_params(params, n_repeats, self._ma_training)

        self._key, reset_keys = get_task_reset_keys(
            self._key, test, self._pop_size, self._n_evaluations, n_repeats,
            self._ma_training)

        # Reset the tasks and the policy.
        task_state = task_reset_func(reset_keys)
        policy_state = policy_reset_func(task_state)
        if self._num_device > 1:
            params = split_params_for_pmap(params)
            task_state = split_states_for_pmap(task_state)
            policy_state = split_states_for_pmap(policy_state)

        # Do the rollouts.
        scores, all_obs, masks = rollout_func(
            task_state, policy_state, params, self.obs_params)
        if self._num_device > 1:
            all_obs = reshape_data_from_pmap(all_obs)
            masks = reshape_data_from_pmap(masks)

        if not test and not self.obs_normalizer.is_dummy:
            self.obs_params = self.obs_normalizer.update_normalization_params(
                obs_buffer=all_obs, obs_mask=masks, obs_params=self.obs_params)

        if self._ma_training:
            if not test:
                # In training, each agent has different parameters.
                return jnp.mean(scores.ravel().reshape((n_repeats, -1)), axis=0)
            else:
                # In tests, they share the same parameters.
                return jnp.mean(scores.ravel().reshape((n_repeats, -1)), axis=1)
        else:
            return jnp.mean(scores.ravel().reshape((-1, n_repeats)), axis=-1)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple
from PIL import Image
from PIL import ImageDraw
import numpy as np

import jax
import jax.numpy as jnp
from jax import random
from flax.struct import dataclass

from evojax.task.base import TaskState
from evojax.task.base import VectorizedTask


GRAVITY = 9.82
CART_MASS = 0.5
POLE_MASS = 0.5
POLE_LEN = 0.6
FRICTION = 0.1
FORCE_SCALING = 10.0
DELTA_T = 0.01
CART_X_LIMIT = 2.4

SCREEN_W = 600
SCREEN_H = 600
CART_W = 40
CART_H = 20
VIZ_SCALE = 100
WHEEL_RAD = 5


@dataclass
class State(TaskState):
    obs: jnp.ndarray
    state: jnp.ndarray
    steps: jnp.int32
    key: jnp.ndarray


def get_init_state_easy(key: jnp.ndarray) -> jnp.ndarray:
    return (random.normal(key, shape=(4,)) * 0.2 +
            jnp.array([0, 0, jnp.pi, 0]))


def get_init_state_hard(key: jnp.ndarray) -> jnp.ndarray:
    return (jnp.multiply(random.uniform(key, shape=(4,)) * 2 - 1,
                         jnp.array([CART_X_LIMIT, 10., jnp.pi / 2., 10.])) +
            jnp.array([0, 0, jnp.pi, 0]))


def get_obs(state: jnp.ndarray) -> jnp.ndarray:
    x, x_dot, theta, theta_dot = state
    return jnp.array([x, x_dot, jnp.cos(theta), jnp.sin(theta), theta_dot])


def get_reward(state: jnp.ndarray) -> jnp.float32:
    x, x_dot, theta, theta_dot = state
    reward_theta = (jnp.cos(theta) + 1.0) / 2.0
    reward_x = jnp.cos((x / CART_X_LIMIT) * (jnp.pi / 2.0))
    return reward_theta * reward_x


def update_state(action: jnp.ndarray, state: jnp.ndarray) -> jnp.ndarray:
    action = jnp.clip(action, -1.0, 1.0)[0] * FORCE_SCALING
    x, x_dot, theta, theta_dot = state
    s = jnp.sin(theta)
    c = jnp.cos(theta)
    total_m = CART_MASS + POLE_MASS
    m_p_l = POLE_MASS * POLE_LEN
    x_dot_update = (
            (-2 * m_p_l * (theta_dot ** 2) * s +
             3 * POLE_MASS * GRAVITY * s * c +
             4 * action - 4 * FRICTION * x_dot) /
            (4 * total_m - 3 * POLE_MASS * c ** 2)
    )
    theta_dot_update = (
            (-3 * m_p_l * (theta_dot ** 2) * s * c +
             6 * total_m * GRAVITY * s +
             6 * (action - FRICTION * x_dot) * c) /
            (4 * POLE_LEN * total_m - 3 * m_p_l * c ** 2)
    )
    x = x + x_dot * DELTA_T
    theta = theta + theta_dot * DELTA_T
    x_dot = x_dot + x_dot_update * DELTA_T
    theta_dot = theta_dot + theta_dot_update * DELTA_T
    return jnp.array([x, x_dot, theta, theta_dot])


def out_of_screen(state: jnp.ndarray) -> jnp.float32:
    x = state[0]
    beyond_boundary_l = jnp.where(x < -CART_X_LIMIT, 1, 0)
    beyond_boundary_r = jnp.where(x > CART_X_LIMIT, 1, 0)
    return jnp.bitwise_or(beyond_boundary_l, beyond_boundary_r)


class CartPoleSwingUp(VectorizedTask):
    """Cart-pole swing up task."""

    def __init__(self,
                 max_steps: int = 1000,
                 harder: bool = False,
                 test: bool = False):

        self.max_steps = max_steps
        self.obs_shape = tuple([5, ])
        self.act_shape = tuple([1, ])
        self.test = test
        if harder:
            get_init_state_fn = get_init_state_hard
        else:
            get_init_state_fn = get_init_state_easy

        def reset_fn(key):
            next_key, key = random.split(key)
            state = get_init_state_fn(key)
            return State(state=state, obs=get_obs(state),
                         steps=jnp.zeros((), dtype=int), key=next_key)
        self._reset_fn = jax.jit(jax.vmap(reset_fn))

        def step_fn(state, action):
            cur_state = update_state(action=action, state=state.state)
            reward = get_reward(state=cur_state)
            steps = state.steps + 1
            done = jnp.bitwise_or(out_of_screen(cur_state), steps >= max_steps)
            steps = jnp.where(done, jnp.zeros((), jnp.int32), steps)
            next_key, key = random.split(state.key)
            cur_state = jax.lax.cond(
                done, lambda x: get_init_state_fn(key), lambda x: x, cur_state)
            return State(state=cur_state, obs=get_obs(state=cur_state),
                         steps=steps, key=next_key), reward, done
        self._step_fn = jax.jit(jax.vmap(step_fn))

    def reset(self, key: jnp.ndarray) -> State:
        return self._reset_fn(key)

    def step(self,
             state: State,
             action: jnp.ndarray) -> Tuple[State, jnp.ndarray, jnp.ndarray]:
        return self._step_fn(state, action)

    @staticmethod
    def render(state: State, task_id: int) -> Image:
        """Render a specified task."""
        img = Image.new('RGB', (SCREEN_W, SCREEN_H), (255, 255, 255))
        draw = ImageDraw.Draw(img)
        x, _, theta, _ = np.array(state.state[task_id])
        cart_y = SCREEN_H // 2 + 100
        cart_x = x * VIZ_SCALE + SCREEN_W // 2
        # Draw the horizon.
        draw.line(
            (0, cart_y + CART_H // 2 + WHEEL_RAD,
             SCREEN_W, cart_y + CART_H // 2 + WHEEL_RAD),
            fill=(0, 0, 0), width=1)
        # Draw the cart.
        draw.rectangle(
            (cart_x - CART_W // 2, cart_y - CART_H // 2,
             cart_x + CART_W // 2, cart_y + CART_H // 2),
            fill=(255, 0, 0), outline=(0, 0, 0))
        # Draw the wheels.
        draw.ellipse(
            (cart_x - CART_W // 2 - WHEEL_RAD,
             cart_y + CART_H // 2 - WHEEL_RAD,
             cart_x - CART_W // 2 + WHEEL_RAD,
             cart_y + CART_H // 2 + WHEEL_RAD),
            fill=(220, 220, 220), outline=(0, 0, 0))
        draw.ellipse(
            (cart_x + CART_W // 2 - WHEEL_RAD,
             cart_y + CART_H // 2 - WHEEL_RAD,
             cart_x + CART_W // 2 + WHEEL_RAD,
             cart_y + CART_H // 2 + WHEEL_RAD),
            fill=(220, 220, 220), outline=(0, 0, 0))
        # Draw the pole.
        draw.line(
            (cart_x, cart_y,
             cart_x + POLE_LEN * VIZ_SCALE * np.cos(theta - np.pi / 2),
             cart_y + POLE_LEN * VIZ_SCALE * np.sin(theta - np.pi / 2)),
            fill=(0, 0, 255), width=6)
        return img



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Tuple
import jax.numpy as jnp


class TaskState(ABC):
    """A template of the task state."""
    obs: jnp.ndarray


class VectorizedTask(ABC):
    """Interface for all the EvoJAX tasks."""

    max_steps: int
    obs_shape: Tuple
    act_shape: Tuple
    test: bool
    multi_agent_training: bool = False

    @abstractmethod
    def reset(self, key: jnp.array) -> TaskState:
        """This resets the vectorized task.

        Args:
            key - A jax random key.
        Returns:
            TaskState. Initial task state.
        """
        raise NotImplementedError()

    @abstractmethod
    def step(self,
             state: TaskState,
             action: jnp.ndarray) -> Tuple[TaskState, jnp.ndarray, jnp.ndarray]:
        """This steps once the simulation.

        Args:
            state - System internal states of shape (num_tasks, *).
            action - Vectorized actions of shape (num_tasks, action_size).
        Returns:
            TaskState. Task states.
            jnp.ndarray. Reward.
            jnp.ndarray. Task termination flag: 1 for done, 0 otherwise.
        """
        raise NotImplementedError()



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Union
import numpy as np
import jax.numpy as jnp


class NEAlgorithm(ABC):
    """Interface of all Neuro-evolution algorithms in EvoJAX."""

    pop_size: int

    @abstractmethod
    def ask(self) -> jnp.ndarray:
        """Ask the algorithm for a population of parameters.

        Returns
            A Jax array of shape (population_size, param_size).
        """
        raise NotImplementedError()

    @abstractmethod
    def tell(self, fitness: Union[np.ndarray, jnp.ndarray]) -> None:
        """Report the fitness of the population to the algorithm.

        Args:
            fitness - The fitness scores array.
        """
        raise NotImplementedError()

    @property
    def best_params(self) -> jnp.ndarray:
        raise NotImplementedError()

    @best_params.setter
    def best_params(self, params: Union[np.ndarray, jnp.ndarray]) -> None:
        raise NotImplementedError()



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Tuple

import jax
import jax.numpy as jnp
from jax import random
from flax import linen as nn

from evojax.policy.base import PolicyNetwork
from evojax.policy.base import PolicyState
from evojax.task.base import TaskState
from evojax.util import create_logger
from evojax.util import get_params_format_fn


class CNN(nn.Module):
    """CNN for MNIST."""

    @nn.compact
    def __call__(self, x):
        x = nn.Conv(features=8, kernel_size=(5, 5), padding='SAME')(x)
        x = nn.relu(x)
        x = nn.max_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = nn.Conv(features=16, kernel_size=(5, 5), padding='SAME')(x)
        x = nn.relu(x)
        x = nn.max_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = x.reshape((x.shape[0], -1))  # flatten
        x = nn.Dense(features=10)(x)
        x = nn.log_softmax(x)
        return x


class ConvNetPolicy(PolicyNetwork):
    """A convolutional neural network for the MNIST classification task."""

    def __init__(self, logger: logging.Logger = None):
        if logger is None:
            self._logger = create_logger('ConvNetPolicy')
        else:
            self._logger = logger

        model = CNN()
        params = model.init(random.PRNGKey(0), jnp.zeros([1, 28, 28, 1]))
        self.num_params, format_params_fn = get_params_format_fn(params)
        self._logger.info(
            'ConvNetPolicy.num_params = {}'.format(self.num_params))
        self._format_params_fn = jax.vmap(format_params_fn)
        self._forward_fn = jax.vmap(model.apply)

    def get_actions(self,
                    t_states: TaskState,
                    params: jnp.ndarray,
                    p_states: PolicyState) -> Tuple[jnp.ndarray, PolicyState]:
        params = self._format_params_fn(params)
        return self._forward_fn(params, t_states.obs), p_states



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the WaterWorld task in JAX.

Ref: https://cs.stanford.edu/people/karpathy/reinforcejs/waterworld.html
"""

from typing import Tuple
from functools import partial
from PIL import Image
from PIL import ImageDraw
import numpy as np

import jax
import jax.numpy as jnp
from jax import random
from jax import tree_util
from flax.struct import dataclass

from evojax.task.base import TaskState
from evojax.task.base import VectorizedTask


SCREEN_W = 600
SCREEN_H = 600
BUBBLE_RADIUS = 10
MIN_DIST = 2 * BUBBLE_RADIUS
MAX_RANGE = 120
NUM_RANGE_SENSORS = 30
DELTA_ANG = 2 * 3.14 / NUM_RANGE_SENSORS

TYPE_VOID = 0
TYPE_WALL = 1
TYPE_FOOD = 2
TYPE_POISON = 3
TYPE_AGENT = 4
SENSOR_DATA_DIM = 5  # dist_wall, dist_food, dist_poison, item_vel_x, item_vel_y

ACT_UP = 0
ACT_DOWN = 1
ACT_LEFT = 2
ACT_RIGHT = 3


@dataclass
class BubbleStatus(object):
    pos_x: jnp.float32
    pos_y: jnp.float32
    vel_x: jnp.float32
    vel_y: jnp.float32
    bubble_type: jnp.int32
    valid: jnp.int32
    poison_cnt: jnp.int32


@dataclass
class State(TaskState):
    agent_state: BubbleStatus
    item_state: BubbleStatus
    obs: jnp.ndarray
    steps: jnp.int32
    key: jnp.ndarray


@partial(jax.vmap, in_axes=(0, None))
def create_bubbles(key: jnp.ndarray, is_agent: bool) -> BubbleStatus:
    k_pos_x, k_pos_y, k_vel, k_bubble_type = random.split(key, 4)
    if is_agent:
        bubble_type = TYPE_AGENT
        vel_x = vel_y = 0.
    else:
        bubble_type = jnp.where(
            random.uniform(k_bubble_type) > 0.5, TYPE_FOOD, TYPE_POISON)
        vel_x, vel_y = random.uniform(
            k_vel, shape=(2,), minval=-2.5, maxval=2.5)
    return BubbleStatus(
        pos_x=random.uniform(
            k_pos_x, shape=(), dtype=jnp.float32,
            minval=MIN_DIST, maxval=SCREEN_W - MIN_DIST),
        pos_y=random.uniform(
            k_pos_y, shape=(), dtype=jnp.float32,
            minval=MIN_DIST, maxval=SCREEN_H - MIN_DIST),
        vel_x=vel_x, vel_y=vel_y, bubble_type=bubble_type,
        valid=1, poison_cnt=0)


def get_reward(agent: BubbleStatus,
               items: BubbleStatus) -> Tuple[BubbleStatus,
                                             BubbleStatus,
                                             jnp.float32]:
    dist = jnp.sqrt(jnp.square(agent.pos_x - items.pos_x) +
                    jnp.square(agent.pos_y - items.pos_y))
    rewards = (jnp.where(items.bubble_type == TYPE_FOOD, 1., -1.) *
               items.valid * jnp.where(dist < MIN_DIST, 1, 0))
    poison_cnt = jnp.sum(jnp.where(rewards == -1., 1, 0)) + agent.poison_cnt
    reward = jnp.sum(rewards)
    items_valid = (dist >= MIN_DIST) * items.valid
    agent_state = BubbleStatus(
        pos_x=agent.pos_x, pos_y=agent.pos_y,
        vel_x=agent.vel_x, vel_y=agent.vel_y,
        bubble_type=agent.bubble_type,
        valid=agent.valid, poison_cnt=poison_cnt)
    items_state = BubbleStatus(
        pos_x=items.pos_x, pos_y=items.pos_y,
        vel_x=items.vel_x, vel_y=items.vel_y,
        bubble_type=items.bubble_type,
        valid=items_valid, poison_cnt=items.poison_cnt)
    return agent_state, items_state, reward


@jax.vmap
def update_item_state(item: BubbleStatus) -> BubbleStatus:
    vel_x = item.vel_x
    vel_y = item.vel_y
    pos_x = item.pos_x + vel_x
    pos_y = item.pos_y + vel_y
    # Collide with the west wall.
    vel_x = jnp.where(pos_x < 1, -vel_x, vel_x)
    pos_x = jnp.where(pos_x < 1, 1, pos_x)
    # Collide with the east wall.
    vel_x = jnp.where(pos_x > SCREEN_W - 1, -vel_x, vel_x)
    pos_x = jnp.where(pos_x > SCREEN_W - 1, SCREEN_W - 1, pos_x)
    # Collide with the north wall.
    vel_y = jnp.where(pos_y < 1, -vel_y, vel_y)
    pos_y = jnp.where(pos_y < 1, 1, pos_y)
    # Collide with the south wall.
    vel_y = jnp.where(pos_y > SCREEN_H - 1, -vel_y, vel_y)
    pos_y = jnp.where(pos_y > SCREEN_H - 1, SCREEN_H - 1, pos_y)
    return BubbleStatus(
        pos_x=pos_x, pos_y=pos_y, vel_x=vel_x, vel_y=vel_y,
        bubble_type=item.bubble_type, valid=item.valid,
        poison_cnt=item.poison_cnt)


def update_agent_state(agent: BubbleStatus,
                       direction: jnp.int32) -> BubbleStatus:
    vel_x = agent.vel_x
    vel_x = jnp.where(direction == ACT_RIGHT, vel_x + 1, vel_x)
    vel_x = jnp.where(direction == ACT_LEFT, vel_x - 1, vel_x)
    vel_x = vel_x * 0.95

    vel_y = agent.vel_y
    vel_y = jnp.where(direction == ACT_UP, vel_y - 1, vel_y)
    vel_y = jnp.where(direction == ACT_DOWN, vel_y + 1, vel_y)
    vel_y = vel_y * 0.95

    pos_x = agent.pos_x + vel_x
    pos_y = agent.pos_y + vel_y
    # Collide with the west wall.
    vel_x = jnp.where(pos_x < 1, 0, vel_x)
    vel_y = jnp.where(pos_x < 1, 0, vel_y)
    pos_x = jnp.where(pos_x < 1, 1, pos_x)
    # Collide with the east wall.
    vel_x = jnp.where(pos_x > SCREEN_W - 1, 0, vel_x)
    vel_y = jnp.where(pos_x > SCREEN_W - 1, 0, vel_y)
    pos_x = jnp.where(pos_x > SCREEN_W - 1, SCREEN_W - 1, pos_x)
    # Collide with the north wall.
    vel_x = jnp.where(pos_y < 1, 0, vel_x)
    vel_y = jnp.where(pos_y < 1, 0, vel_y)
    pos_y = jnp.where(pos_y < 1, 1, pos_y)
    # Collide with the south wall.
    vel_x = jnp.where(pos_y > SCREEN_H - 1, 0, vel_x)
    vel_y = jnp.where(pos_y > SCREEN_H - 1, 0, vel_y)
    pos_y = jnp.where(pos_y > SCREEN_H - 1, SCREEN_H - 1, pos_y)

    return BubbleStatus(
        pos_x=pos_x, pos_y=pos_y, vel_x=vel_x, vel_y=vel_y,
        bubble_type=agent.bubble_type, valid=agent.valid,
        poison_cnt=agent.poison_cnt)


@jax.vmap
def get_line_seg_intersection(x1: jnp.float32,
                              y1: jnp.float32,
                              x2: jnp.float32,
                              y2: jnp.float32,
                              x3: jnp.float32,
                              y3: jnp.float32,
                              x4: jnp.float32,
                              y4: jnp.float32) -> Tuple[bool, jnp.ndarray]:
    """Determine if line segment (x1, y1, x2, y2) intersects with line
    segment (x3, y3, x4, y4), and return the intersection coordinate.
    """
    denominator = (y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1)
    ua = jnp.where(
        jnp.isclose(denominator, 0.0), 0,
        ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / denominator)
    mask1 = jnp.bitwise_and(ua > 0., ua < 1.)
    ub = jnp.where(
        jnp.isclose(denominator, 0.0), 0,
        ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) / denominator)
    mask2 = jnp.bitwise_and(ub > 0., ub < 1.)
    intersected = jnp.bitwise_and(mask1, mask2)
    x_intersection = x1 + ua * (x2 - x1)
    y_intersection = y1 + ua * (y2 - y1)
    up = jnp.where(intersected,
                   jnp.array([x_intersection, y_intersection]),
                   jnp.array([SCREEN_W, SCREEN_W]))
    return intersected, up


@jax.vmap
def get_line_dot_intersection(x1: jnp.float32,
                              y1: jnp.float32,
                              x2: jnp.float32,
                              y2: jnp.float32,
                              x3: jnp.float32,
                              y3: jnp.float32) -> Tuple[bool, jnp.ndarray]:
    """Determine if a line segment (x1, y1, x2, y2) intersects with a dot at
    (x3, y3) with radius BUBBLE_RADIUS, if so return the point of intersection.
    """
    point_xy = jnp.array([x3, y3])
    v = jnp.array([y2 - y1, x1 - x2])
    v_len = jnp.linalg.norm(v)
    d = jnp.abs((x2 - x1) * (y1 - y3) - (x1 - x3) * (y2 - y1)) / v_len
    up = point_xy + v / v_len * d
    ua = jnp.where(jnp.abs(x2 - x1) > jnp.abs(y2 - y1),
                   (up[0] - x1) / (x2 - x1),
                   (up[1] - y1) / (y2 - y1))
    ua = jnp.where(d > BUBBLE_RADIUS, 0, ua)
    intersected = jnp.bitwise_and(ua > 0., ua < 1.)
    return intersected, up


def get_obs(agent: BubbleStatus,
            items: BubbleStatus,
            walls: jnp.ndarray) -> jnp.ndarray:
    sensor_obs = []
    agent_xy = jnp.array([agent.pos_x, agent.pos_y]).ravel()
    for i in jnp.arange(NUM_RANGE_SENSORS):
        ang = i * DELTA_ANG
        range_xy = jnp.array([agent.pos_x + MAX_RANGE * jnp.cos(ang),
                              agent.pos_y + MAX_RANGE * jnp.sin(ang)])
        # Check for intersections with the 4 walls.
        intersected_with_walls, wall_intersections = get_line_seg_intersection(
            jnp.ones(4) * agent_xy[0], jnp.ones(4) * agent_xy[1],
            jnp.ones(4) * range_xy[0], jnp.ones(4) * range_xy[1],
            walls[:, 0], walls[:, 1], walls[:, 2], walls[:, 3])
        dist_to_walls = jnp.where(
            intersected_with_walls,
            jnp.sqrt(jnp.square(wall_intersections[:, 1] - agent_xy[1]) +
                     jnp.square(wall_intersections[:, 0] - agent_xy[0])),
            MAX_RANGE,
        )
        ix_walls = jnp.argmin(dist_to_walls)
        # Check for intersections with the items.
        n_items = len(items.valid)
        intersected_with_items, item_intersections = get_line_dot_intersection(
            jnp.ones(n_items) * agent_xy[0], jnp.ones(n_items) * agent_xy[1],
            jnp.ones(n_items) * range_xy[0], jnp.ones(n_items) * range_xy[1],
            items.pos_x, items.pos_y)
        dist_to_items = jnp.where(
            jnp.bitwise_and(items.valid, intersected_with_items),
            jnp.sqrt(jnp.square(item_intersections[:, 1] - agent_xy[1]) +
                     jnp.square(item_intersections[:, 0] - agent_xy[0])),
            MAX_RANGE,
        )
        ix_items = jnp.argmin(dist_to_items)
        # Fill in the sensor data.
        detected_xy = jnp.where(
            jnp.min(dist_to_walls) < jnp.min(dist_to_items),
            jnp.where(
                intersected_with_walls[ix_walls],
                jnp.array([
                    dist_to_walls[ix_walls], MAX_RANGE, MAX_RANGE, 0, 0]),
                jnp.array([
                    MAX_RANGE, MAX_RANGE, MAX_RANGE, 0, 0])),
            jnp.where(
                intersected_with_items[ix_items],
                jnp.where(
                    items.bubble_type[ix_items] == TYPE_FOOD,
                    jnp.array([
                        MAX_RANGE, dist_to_items[ix_items], MAX_RANGE,
                        items.vel_x[ix_items], items.vel_y[ix_items]]),
                    jnp.array([
                        MAX_RANGE, MAX_RANGE, dist_to_items[ix_items],
                        items.vel_x[ix_items], items.vel_y[ix_items]]),
                ),
                jnp.array([
                    MAX_RANGE, MAX_RANGE, MAX_RANGE, 0, 0])),
            )
        sensor_obs.append(detected_xy)

    sensor_obs = jnp.stack(sensor_obs)
    sensor_obs = sensor_obs.at[:, :3].divide(MAX_RANGE)  # Normalized distances.
    vel_xy = jnp.array([agent.vel_x, agent.vel_y]).ravel()
    return jnp.concatenate([sensor_obs.ravel(), vel_xy], axis=0)


class WaterWorld(VectorizedTask):
    """Water world.
    ref: https://cs.stanford.edu/people/karpathy/reinforcejs/waterworld.html
    """

    def __init__(self,
                 num_items: int = 50,
                 max_steps: int = 1000,
                 test: bool = False):

        self.max_steps = max_steps
        self.test = test
        self.obs_shape = tuple([NUM_RANGE_SENSORS * SENSOR_DATA_DIM + 2, ])
        self.act_shape = tuple([4, ])
        walls = jnp.array([[0, 0, 0, SCREEN_H],
                           [0, SCREEN_H, SCREEN_W, SCREEN_H],
                           [SCREEN_W, SCREEN_H, SCREEN_W, 0],
                           [SCREEN_W, 0, 0, 0]])

        def reset_fn(key):
            next_key, key = random.split(key)
            ks = random.split(key, 1 + num_items)
            agent = create_bubbles(ks[0][None, :], True)
            items = create_bubbles(ks[1:], False)
            obs = get_obs(agent, items, walls)
            return State(agent_state=agent, item_state=items, obs=obs,
                         steps=jnp.zeros((), dtype=jnp.int32), key=next_key)
        self._reset_fn = jax.jit(jax.vmap(reset_fn))

        def step_fn(state, action):
            next_key, key = random.split(state.key)
            direction = random.choice(
                key, 4, (), replace=False, p=action.ravel())
            agent = update_agent_state(state.agent_state, direction)
            items = update_item_state(state.item_state)
            agent, items, reward = get_reward(agent, items)
            steps = state.steps + 1
            done = jnp.where(steps >= max_steps, 1, 0)
            obs = get_obs(agent, items, walls)
            return State(agent_state=agent, item_state=items, obs=obs,
                         steps=steps, key=next_key), reward, done
        self._step_fn = jax.jit(jax.vmap(step_fn))

    def reset(self, key: jnp.array) -> State:
        return self._reset_fn(key)

    def step(self,
             state: State,
             action: jnp.ndarray) -> Tuple[State, jnp.ndarray, jnp.ndarray]:
        return self._step_fn(state, action)

    @staticmethod
    def render(state: State, task_id: int = 0) -> Image:
        img = Image.new('RGB', (SCREEN_W, SCREEN_H), (255, 255, 255))
        draw = ImageDraw.Draw(img)
        state = tree_util.tree_map(lambda s: s[task_id], state)
        # Draw the agent.
        agent = state.agent_state
        x, y = agent.pos_x, agent.pos_y
        sensor_data = np.array(state.obs)[:-2].reshape(
            NUM_RANGE_SENSORS, SENSOR_DATA_DIM)
        for i, obs in enumerate(sensor_data):
            ang = i * DELTA_ANG
            dist = np.min(obs[:3])
            x_end = x + dist * MAX_RANGE * np.cos(ang)
            y_end = y + dist * MAX_RANGE * np.sin(ang)
            draw.line((x, y, x_end, y_end), fill=(0, 0, 0), width=1)
        draw.ellipse(
            (x - BUBBLE_RADIUS, y - BUBBLE_RADIUS,
             x + BUBBLE_RADIUS, y + BUBBLE_RADIUS),
            fill=(255, 255, 0), outline=(0, 0, 0))
        # Draw the items.
        items = state.item_state
        for v, t, x, y in zip(np.array(items.valid, dtype=bool),
                              np.array(items.bubble_type, dtype=int),
                              np.array(items.pos_x),
                              np.array(items.pos_y)):
            if v:
                color = (0, 255, 0) if t == TYPE_FOOD else (255, 0, 0)
                draw.ellipse(
                    (x - BUBBLE_RADIUS, y - BUBBLE_RADIUS,
                     x + BUBBLE_RADIUS, y + BUBBLE_RADIUS,),
                    fill=color, outline=(0, 0, 0))
        return img



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image, ImageDraw
from functools import partial
from typing import Tuple

import jax
from jax import vmap
import jax.numpy as jnp
from flax.struct import dataclass

from evojax.task.base import VectorizedTask
from evojax.task.base import TaskState

DT = 0.2
SPEED = 0.12
J_ALIGN = 0.6
D_ALIGN = 0.25
J_AVOID = 0.12
D_AVOID = 0.08
J_COHESION = 0.025
D_COHESION = 0.2
ALPHA = 3

SCREEN_W = 400
SCREEN_H = 300
FISH_SIZE = 20.
BOIDS_NUM = 30
NEIGHBOR_NUM = 5


@dataclass
class State(TaskState):
    obs: jnp.ndarray
    state: jnp.ndarray
    steps: jnp.int32
    key: jnp.ndarray


def sample_position(key: jnp.ndarray, n: jnp.ndarray) -> jnp.ndarray:
    return jax.random.uniform(key, shape=(n, 2), minval=0.0, maxval=1.0)


def sample_theta(key: jnp.ndarray, n: jnp.ndarray) -> jnp.ndarray:
    return jax.random.uniform(key, shape=(n, 1), maxval=2. * jnp.pi)


def unpack_obs(obs: jnp.ndarray) -> jnp.ndarray:
    position, theta = obs[..., :2], obs[..., 2:]
    return position, theta


def pack_obs(position: jnp.ndarray, theta: jnp.ndarray) -> jnp.ndarray:
    return jnp.concatenate([position, theta], axis=-1)


def displacement(p1: jnp.ndarray, p2: jnp.ndarray) -> jnp.ndarray:
    dR = p1 - p2
    return jnp.mod(dR + 0.5, 1) - 0.5


def map_product(displacement):
    return vmap(vmap(displacement, (0, None), 0), (None, 0), 0)


def calc_distance(dR: jnp.ndarray) -> jnp.ndarray:
    dr = jnp.sqrt(jnp.sum(dR**2, axis=-1))
    return dr


def select_xy(xy: jnp.ndarray, ix: jnp.ndarray) -> jnp.ndarray:
    return jnp.take(xy, ix, axis=0)


_select_xy = jax.vmap(select_xy, in_axes=(None, 0))


def choose_neighbor(state: jnp.ndarray) -> jnp.ndarray:
    position, _ = unpack_obs(state)
    dR = map_product(displacement)(position, position)
    dr = calc_distance(dR)
    sort_idx = jnp.argsort(dr, axis=1)
    neighbor_ix = sort_idx[:, :NEIGHBOR_NUM]
    neighbors = _select_xy(state, neighbor_ix)
    return jnp.reshape(neighbors, (len(state), -1))


def normal(theta: jnp.ndarray) -> jnp.ndarray:
    return jnp.concatenate([jnp.cos(theta), jnp.sin(theta)], axis=-1)


def field_of_view_mask(dR, N, theta_min, theta_max):
    dr = calc_distance(dR)
    dR_hat = dR / dr
    ctheta = jnp.dot(dR_hat, N)
    return jnp.logical_and(ctheta > jnp.cos(theta_max),
                           ctheta < jnp.cos(theta_min))


def align_fn(dR, N_1, N_2, J_align, D_align, alpha):
    dr = calc_distance(dR) / D_align
    energy = J_align / alpha * (1. - dr)**alpha * (1 - jnp.dot(N_1, N_2))**2
    return jnp.where(dr < 1.0, energy, 0.)


def avoid_fn(dR, J_avoid, D_avoid, alpha):
    dr = calc_distance(dR) / D_avoid
    energy = J_avoid / alpha * (1. - dr)**alpha
    return jnp.where(dr < 1.0, energy, 0.)


def cohesion_fn(dR, N, mask, J_cohesion, D_cohesion, eps=1e-7):
    dr = calc_distance(dR)

    mask = jnp.reshape(mask, mask.shape + (1, ))
    dr = jnp.reshape(dr, dr.shape + (1, ))
    mask = jnp.logical_and(dr < D_cohesion, mask)

    N_com = jnp.where(mask, 1.0, 0)
    dR_com = jnp.where(mask, dR, 0)
    dR_com = jnp.sum(dR_com, axis=1) / (jnp.sum(N_com, axis=1) + eps)
    dR_com = dR_com / jnp.linalg.norm(dR_com + eps, axis=1, keepdims=True)
    return 0.5 * J_cohesion * (1 - jnp.sum(dR_com * N, axis=1))**2


def calc_energy(position, theta):
    E_align = partial(align_fn, J_align=J_ALIGN, D_align=D_ALIGN, alpha=ALPHA)
    E_align = vmap(vmap(E_align, (0, None, 0)), (0, 0, None))

    E_avoid = partial(avoid_fn, J_avoid=J_AVOID, D_avoid=D_AVOID, alpha=ALPHA)
    E_avoid = vmap(vmap(E_avoid))

    E_cohesion = partial(cohesion_fn,
                         J_cohesion=J_COHESION,
                         D_cohesion=D_COHESION)

    dR = map_product(displacement)(position, position)
    N = normal(theta)

    fov = partial(field_of_view_mask, theta_min=0., theta_max=jnp.pi / 3.)
    fov = vmap(vmap(fov, (0, None)))
    mask = fov(dR, N)

    return 0.5 * jnp.sum(E_align(dR, N, N) + E_avoid(dR)) + jnp.sum(
        E_cohesion(dR, N, mask))


def update_state(state, action):
    position, theta = unpack_obs(state.state)
    d_theta = action
    N = normal(theta)
    new_obs = pack_obs(jnp.mod(position + DT * SPEED * N, 1),
                       theta + DT * d_theta)
    return new_obs


def get_reward(state: State, max_steps: jnp.int32):
    position, theta = unpack_obs(state.state)
    reward = -calc_energy(position, theta)
    gamma = state.steps / max_steps
    return reward * gamma ** 2


def to_pillow_coordinate(position, width, height):
    # Fix the format for drawing with pillow.
    return jnp.stack([position[:, 0] * width,
                      (1.0 - position[:, 1]) * height], axis=1)


def rotate(px, py, cx, cy, angle):
    R = jnp.array([[jnp.cos(angle), jnp.sin(angle)],
                   [-jnp.sin(angle), jnp.cos(angle)]])
    u = jnp.array([px - cx, py - cy])
    x, y = jnp.dot(R, u) + jnp.array([cx, cy])
    return x, y


def render_fishes(position, theta, width, height):
    position = to_pillow_coordinate(position, width, height)
    colors = [(88, 167, 210), (0, 153, 255), (56, 96, 123)]
    color_map = [i % len(colors) for i in range(BOIDS_NUM)]

    image = Image.new("RGB", (width, height), (255, 255, 255))
    draw = ImageDraw.Draw(image)

    fish_size = FISH_SIZE
    fish_size_half = (fish_size + 1) // 2

    for i in range(BOIDS_NUM):
        x, y = position[i]
        angle = theta[i][0]

        polygon_xy = [
            (x - fish_size_half, y - fish_size_half / 3),
            (x - fish_size_half / 5 * 3, y - fish_size_half / 4),
            (x, y - fish_size_half / 2),
            (x + fish_size_half / 3 * 2, y - fish_size_half / 3),
            (x + fish_size_half, y),
            (x + fish_size_half / 3 * 2, y + fish_size_half / 3),
            (x, y + fish_size_half / 2),
            (x - fish_size_half / 5 * 3, y + fish_size_half / 4),
            (x - fish_size_half, y + fish_size_half / 3),
        ]

        polygon_xy = [rotate(px, py, x, y, angle) for px, py in polygon_xy]
        draw.polygon(xy=polygon_xy, fill=colors[color_map[i]])
    return image


def render_single(obs_single):
    position, theta = unpack_obs(obs_single)
    image = render_fishes(position, theta, SCREEN_W * 2, SCREEN_H * 2)  # anti-aliasing
    image = image.resize((SCREEN_W, SCREEN_H), resample=Image.LANCZOS)
    return image


class FlockingTask(VectorizedTask):

    def __init__(self, max_steps: int = 150):
        self.max_steps = max_steps
        self.obs_shape = tuple([NEIGHBOR_NUM * 3, BOIDS_NUM])
        self.act_shape = tuple([1, ])

        def reset_fn(key):
            next_key, key = jax.random.split(key)
            position = sample_position(key, BOIDS_NUM)
            theta = sample_theta(key, BOIDS_NUM)
            state = pack_obs(position, theta)
            return State(obs=choose_neighbor(state),
                         state=state,
                         steps=jnp.zeros((), dtype=jnp.int32),
                         key=next_key)
        self._reset_fn = jax.jit(jax.vmap(reset_fn))

        def step_fn(state, action):
            new_state = update_state(state, action)
            new_obs = choose_neighbor(new_state)
            new_steps = jnp.int32(state.steps + 1)
            next_key, _ = jax.random.split(state.key)
            reward = get_reward(state, max_steps)
            done = jnp.where(max_steps <= new_steps, True, False)
            return State(obs=new_obs,
                         state=new_state,
                         steps=new_steps,
                         key=next_key), reward, done
        self._step_fn = jax.jit(jax.vmap(step_fn))

    def step(self, state: State,
             action: jnp.ndarray) -> Tuple[State, jnp.ndarray, jnp.ndarray]:
        return self._step_fn(state, action)

    def reset(self, key: jnp.ndarray):
        return self._reset_fn(key)

    @staticmethod
    def render(state: State, task_id: int) -> Image:
        return render_single(state.state[task_id])



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from typing import Tuple

import jax
import jax.numpy as jnp
from flax.struct import dataclass

from evojax.task.base import VectorizedTask
from evojax.task.base import TaskState

try:
    from brax.envs import create
    from brax.envs import State as BraxState
except ModuleNotFoundError:
    print('You need to install brax for Brax tasks:')
    print('  pip install git+https://github.com/google/brax.git@main')
    sys.exit(1)


@dataclass
class State(TaskState):
    state: BraxState
    obs: jnp.ndarray


class BraxTask(VectorizedTask):
    """Tasks from the Brax simulator."""

    def __init__(self,
                 env_name: str,
                 max_steps: int = 1000,
                 test: bool = False):
        self.max_steps = max_steps
        self.test = test
        brax_env = create(env_name=env_name, episode_length=max_steps)
        self.obs_shape = tuple([brax_env.observation_size, ])
        self.act_shape = tuple([brax_env.action_size, ])

        def reset_fn(key):
            state = brax_env.reset(key)
            return State(state=state, obs=state.obs)

        self._reset_fn = jax.jit(jax.vmap(reset_fn))

        def step_fn(state, action):
            state = brax_env.step(state.state, action)
            return State(state=state, obs=state.obs), state.reward, state.done

        self._step_fn = jax.jit(jax.vmap(step_fn))

    def reset(self, key: jnp.ndarray) -> State:
        return self._reset_fn(key)

    def step(self,
             state: State,
             action: jnp.ndarray) -> Tuple[State, jnp.ndarray, jnp.ndarray]:
        return self._step_fn(state, action)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from typing import Tuple
from functools import partial

import jax
import jax.numpy as jnp


def normalize(obs: jnp.ndarray,
              obs_params: jnp.ndarray,
              obs_shape: Tuple,
              clip_value: float,
              std_min_value: float,
              std_max_value: float) -> jnp.ndarray:
    """Normalize the given observation."""

    obs_steps = obs_params[0]
    running_mean, running_var = jnp.split(obs_params[1:], 2)
    running_mean = running_mean.reshape(obs_shape)
    running_var = running_var.reshape(obs_shape)

    variance = running_var / (obs_steps + 1.0)
    variance = jnp.clip(variance, std_min_value, std_max_value)
    return jnp.clip(
        (obs - running_mean) / jnp.sqrt(variance), -clip_value, clip_value)


def update_obs_params(obs_buffer: jnp.ndarray,
                      obs_mask: jnp.ndarray,
                      obs_params: jnp.ndarray) -> jnp.ndarray:
    """Update observation normalization parameters."""

    obs_steps = obs_params[0]
    running_mean, running_var = jnp.split(obs_params[1:], 2)
    if obs_mask.ndim != obs_buffer.ndim:
        obs_mask = obs_mask.reshape(
            obs_mask.shape + (1,) * (obs_buffer.ndim - obs_mask.ndim))

    new_steps = jnp.sum(obs_mask)
    total_steps = obs_steps + new_steps

    input_to_old_mean = (obs_buffer - running_mean) * obs_mask
    mean_diff = jnp.sum(input_to_old_mean / total_steps, axis=(0, 1))
    new_mean = running_mean + mean_diff

    input_to_new_mean = (obs_buffer - new_mean) * obs_mask
    var_diff = jnp.sum(input_to_new_mean * input_to_old_mean, axis=(0, 1))
    new_var = running_var + var_diff

    return jnp.concatenate([jnp.ones(1) * total_steps, new_mean, new_var])


class ObsNormalizer(object):
    """Observation normalizer."""

    def __init__(self,
                 obs_shape: Tuple,
                 clip_value: float = 5.,
                 std_min_value: float = 1e-6,
                 std_max_value: float = 1e6,
                 dummy: bool = False):
        """Initialization.

        Args:
            obs_shape - Shape of the observations.
            std_min_value - Minimum standard deviation.
            std_max_value - Maximum standard deviation.
            dummy - Whether this is a dummy normalizer.
        """

        self._obs_shape = obs_shape
        self._obs_size = np.prod(obs_shape)
        self._std_min_value = std_min_value
        self._std_max_value = std_max_value
        self._clip_value = clip_value
        self.is_dummy = dummy

    @partial(jax.jit, static_argnums=(0,))
    def normalize_obs(self,
                      obs: jnp.ndarray,
                      obs_params: jnp.ndarray) -> jnp.ndarray:
        """Normalize the given observation.

        Args:
            obs - The observation to be normalized.
        Returns:
            Normalized observation.
        """

        if self.is_dummy:
            return obs
        else:
            return normalize(
                obs=obs,
                obs_params=obs_params,
                obs_shape=self._obs_shape,
                clip_value=self._clip_value,
                std_min_value=self._std_min_value,
                std_max_value=self._std_max_value)

    @partial(jax.jit, static_argnums=(0,))
    def update_normalization_params(self,
                                    obs_buffer: jnp.ndarray,
                                    obs_mask: jnp.ndarray,
                                    obs_params: jnp.ndarray) -> jnp.ndarray:
        """Update internal parameters."""

        if self.is_dummy:
            return jnp.zeros_like(obs_params)
        else:
            return update_obs_params(
                obs_buffer=obs_buffer,
                obs_mask=obs_mask,
                obs_params=obs_params,
            )

    @partial(jax.jit, static_argnums=(0,))
    def get_init_params(self) -> jnp.ndarray:
        return jnp.zeros(1 + self._obs_size * 2)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Tuple
import jax.numpy as jnp
import jax.random
from flax.struct import dataclass

from evojax.task.base import TaskState


@dataclass
class PolicyState(object):
    """Policy internal states."""

    keys: jnp.ndarray


class PolicyNetwork(ABC):
    """Interface for all policy networks in EvoJAX."""

    num_params: int

    def reset(self, states: TaskState) -> PolicyState:
        """Reset the policy.

        Args:
            TaskState - Initial observations.
        Returns:
            PolicyState. Policy internal states.
        """
        keys = jax.random.split(jax.random.PRNGKey(0), states.obs.shape[0])
        return PolicyState(keys=keys)

    @abstractmethod
    def get_actions(self,
                    t_states: TaskState,
                    params: jnp.ndarray,
                    p_states: PolicyState) -> Tuple[jnp.ndarray, PolicyState]:
        """Get vectorized actions.

        Args:
            t_states - Task states.
            params - A batch of parameters, shape is (num_envs, param_size).
            p_states - Policy internal states.
        Returns:
            jnp.ndarray. Vectorized actions.
            PolicyState. Internal states.
        """
        raise NotImplementedError()



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup

_dct = {}
with open("evojax/version.py") as f:
    exec(f.read(), _dct)
__version__ = _dct["__version__"]

JAX_URL = "https://storage.googleapis.com/jax-releases/jax_releases.html"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="evojax",
    version=__version__,
    author="Google",
    author_email="evojax-dev@google.com",
    description="EvoJAX: Hardware-accelerated Neuroevolution.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/google/evojax",
    license="Apache 2.0",
    packages=[
        package for package in find_packages() if package.startswith("evojax")
    ],
    zip_safe=False,
    install_requires=[
        "flax",
        "jax>=0.2.17",
        "jaxlib>=0.1.65",
        "Pillow",
        "cma",
    ],
    extras_require={
        "extra": ['evosax', 'torchvision'],
    },
    dependency_links=[JAX_URL],
    python_requires=">=3.6",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import numpy as np

from evojax.task.base import VectorizedTask
from evojax.policy import PolicyNetwork
from evojax.algo import NEAlgorithm
from evojax.sim_mgr import SimManager
from evojax.obs_norm import ObsNormalizer
from evojax.util import create_logger
from evojax.util import load_model
from evojax.util import save_model


class Trainer(object):
    """A trainer that organizes the training logistics."""

    def __init__(self,
                 policy: PolicyNetwork,
                 solver: NEAlgorithm,
                 train_task: VectorizedTask,
                 test_task: VectorizedTask,
                 max_iter: int = 1000,
                 log_interval: int = 20,
                 test_interval: int = 100,
                 n_repeats: int = 1,
                 test_n_repeats: int = 1,
                 n_evaluations: int = 100,
                 seed: int = 42,
                 debug: bool = False,
                 normalize_obs: bool = False,
                 model_dir: str = None,
                 log_dir: str = None,
                 logger: logging.Logger = None):
        """Initialization.

        Args:
            policy - The policy network to use.
            solver - The ES algorithm for optimization.
            train_task - The task for training.
            test_task - The task for evaluation.
            max_iter - Maximum number of training iterations.
            log_interval - Interval for logging.
            test_interval - Interval for tests.
            n_repeats - Number of rollout repetitions.
            n_evaluations - Number of tests to conduct.
            seed - Random seed to use.
            debug - Whether to turn on the debug flag.
            normalize_obs - Whether to use an observation normalizer.
            model_dir - Directory to save/load model.
            log_dir - Directory to dump logs.
            logger - Logger.
        """

        if logger is None:
            self._logger = create_logger(
                name='Trainer', log_dir=log_dir, debug=debug)
        else:
            self._logger = logger

        self._log_interval = log_interval
        self._test_interval = test_interval
        self._max_iter = max_iter
        self.model_dir = model_dir
        self._log_dir = log_dir

        self._obs_normalizer = ObsNormalizer(
            obs_shape=train_task.obs_shape,
            dummy=not normalize_obs,
        )

        self.solver = solver
        self.sim_mgr = SimManager(
            n_repeats=n_repeats,
            test_n_repeats=test_n_repeats,
            pop_size=solver.pop_size,
            n_evaluations=n_evaluations,
            policy_net=policy,
            train_vec_task=train_task,
            valid_vec_task=test_task,
            seed=seed,
            obs_normalizer=self._obs_normalizer,
            logger=self._logger,
        )

    def run(self, demo_mode: bool = False) -> float:
        """Start the training / test process."""

        if self.model_dir is not None:
            params, obs_params = load_model(model_dir=self.model_dir)
            self.sim_mgr.obs_params = obs_params
            self._logger.info(
                'Loaded model parameters from {}.'.format(self.model_dir))
        else:
            params = None

        if demo_mode:
            if params is None:
                raise ValueError('No policy parameters to evaluate.')
            self._logger.info('Start to test the parameters.')
            scores = np.array(
                self.sim_mgr.eval_params(params=params, test=True))
            self._logger.info(
                '[TEST] #tests={0}, max={1:.4f}, avg={2:.4f}, min={3:.4f}, '
                'std={4:.4f}'.format(scores.size, scores.max(), scores.mean(),
                                     scores.min(), scores.std()))
            return scores.mean()
        else:
            self._logger.info(
                'Start to train for {} iterations.'.format(self._max_iter))

            if params is not None:
                # Continue training from the breakpoint.
                self.solver.best_params = params

            best_score = -float('Inf')

            for i in range(self._max_iter):
                start_time = time.perf_counter()
                params = self.solver.ask()
                self._logger.debug('solver.ask time: {0:.4f}s'.format(
                    time.perf_counter() - start_time))

                start_time = time.perf_counter()
                scores = self.sim_mgr.eval_params(params=params, test=False)
                self._logger.debug('sim_mgr.eval_params time: {0:.4f}s'.format(
                    time.perf_counter() - start_time))

                start_time = time.perf_counter()
                self.solver.tell(fitness=scores)
                self._logger.debug('solver.tell time: {0:.4f}s'.format(
                    time.perf_counter() - start_time))

                if i > 0 and i % self._log_interval == 0:
                    scores = np.array(scores)
                    self._logger.info(
                        'Iter={0}, size={1}, max={2:.4f}, '
                        'avg={3:.4f}, min={4:.4f}, std={5:.4f}'.format(
                            i, scores.size, scores.max(), scores.mean(),
                            scores.min(), scores.std()))

                if i > 0 and i % self._test_interval == 0:
                    best_params = self.solver.best_params
                    test_scores = self.sim_mgr.eval_params(
                        params=best_params, test=True)
                    self._logger.info(
                        '[TEST] Iter={0}, #tests={1}, max={2:.4f} avg={3:.4f}, '
                        'min={4:.4f}, std={5:.4f}'.format(
                            i, test_scores.size, test_scores.max(),
                            test_scores.mean(), test_scores.min(),
                            test_scores.std()))
                    mean_test_score = test_scores.mean()
                    save_model(
                        model_dir=self._log_dir,
                        model_name='iter_{}'.format(i),
                        params=best_params,
                        obs_params=self.sim_mgr.obs_params,
                        best=mean_test_score > best_score,
                    )
                    best_score = max(best_score, mean_test_score)

            # Test and save the final model.
            best_params = self.solver.best_params
            test_scores = self.sim_mgr.eval_params(
                params=best_params, test=True)
            self._logger.info(
                '[TEST] Iter={0}, #tests={1}, max={2:.4f}, avg={3:.4f}, '
                'min={4:.4f}, std={5:.4f}'.format(
                    self._max_iter, test_scores.size, test_scores.max(),
                    test_scores.mean(), test_scores.min(), test_scores.std()))
            mean_test_score = test_scores.mean()
            save_model(
                model_dir=self._log_dir,
                model_name='final',
                params=best_params,
                obs_params=self.sim_mgr.obs_params,
                best=mean_test_score > best_score,
            )
            best_score = max(best_score, mean_test_score)
            self._logger.info(
                'Training done, best_score={0:.4f}'.format(best_score))

            return best_score
